@@ -2,7 +2,9 @@
 // SPDX-License-Identifier: 0BSD
 
 #include "dstring.h"
-#include <assert.h>
+#include "alloc.h"
+#include "bit.h"
+#include "diag.h"
 #include <stdarg.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -24,7 +26,7 @@ static struct dstring *dstrheader(const char *dstr) {
 
 /** Get the correct size for a dstring with the given capacity. */
 static size_t dstrsize(size_t capacity) {
-	return BFS_FLEX_SIZEOF(struct dstring, data, capacity + 1);
+	return sizeof_flex(struct dstring, data, capacity + 1);
 }
 
 /** Allocate a dstring with the given contents. */
@@ -52,12 +54,18 @@ char *dstralloc(size_t capacity) {
 }
 
 char *dstrdup(const char *str) {
-	size_t len = strlen(str);
-	return dstralloc_impl(len, len, str);
+	return dstrxdup(str, strlen(str));
 }
 
 char *dstrndup(const char *str, size_t n) {
-	size_t len = strnlen(str, n);
+	return dstrxdup(str, strnlen(str, n));
+}
+
+char *dstrddup(const char *dstr) {
+	return dstrxdup(dstr, dstrlen(dstr));
+}
+
+char *dstrxdup(const char *str, size_t len) {
 	return dstralloc_impl(len, len, str);
 }
 
@@ -66,10 +74,15 @@ size_t dstrlen(const char *dstr) {
 }
 
 int dstreserve(char **dstr, size_t capacity) {
+	if (!*dstr) {
+		*dstr = dstralloc(capacity);
+		return *dstr ? 0 : -1;
+	}
+
 	struct dstring *header = dstrheader(*dstr);
 
 	if (capacity > header->capacity) {
-		capacity *= 2;
+		capacity = bit_ceil(capacity + 1) - 1;
 
 		header = realloc(header, dstrsize(capacity));
 		if (!header) {
@@ -91,37 +104,56 @@ int dstresize(char **dstr, size_t length) {
 	struct dstring *header = dstrheader(*dstr);
 	header->length = length;
 	header->data[length] = '\0';
-
 	return 0;
 }
 
-/** Common implementation of dstr{cat,ncat,app}. */
-static int dstrcat_impl(char **dest, const char *src, size_t srclen) {
+int dstrcat(char **dest, const char *src) {
+	return dstrxcat(dest, src, strlen(src));
+}
+
+int dstrncat(char **dest, const char *src, size_t n) {
+	return dstrxcat(dest, src, strnlen(src, n));
+}
+
+int dstrdcat(char **dest, const char *src) {
+	return dstrxcat(dest, src, dstrlen(src));
+}
+
+int dstrxcat(char **dest, const char *src, size_t len) {
 	size_t oldlen = dstrlen(*dest);
-	size_t newlen = oldlen + srclen;
+	size_t newlen = oldlen + len;
 
 	if (dstresize(dest, newlen) != 0) {
 		return -1;
 	}
 
-	memcpy(*dest + oldlen, src, srclen);
+	memcpy(*dest + oldlen, src, len);
 	return 0;
 }
 
-int dstrcat(char **dest, const char *src) {
-	return dstrcat_impl(dest, src, strlen(src));
-}
-
-int dstrncat(char **dest, const char *src, size_t n) {
-	return dstrcat_impl(dest, src, strnlen(src, n));
-}
-
-int dstrdcat(char **dest, const char *src) {
-	return dstrcat_impl(dest, src, dstrlen(src));
-}
-
 int dstrapp(char **str, char c) {
-	return dstrcat_impl(str, &c, 1);
+	return dstrxcat(str, &c, 1);
+}
+
+int dstrcpy(char **dest, const char *src) {
+	return dstrxcpy(dest, src, strlen(src));
+}
+
+int dstrncpy(char **dest, const char *src, size_t n) {
+	return dstrxcpy(dest, src, strnlen(src, n));
+}
+
+int dstrdcpy(char **dest, const char *src) {
+	return dstrxcpy(dest, src, dstrlen(src));
+}
+
+int dstrxcpy(char **dest, const char *src, size_t len) {
+	if (dstresize(dest, len) != 0) {
+		return -1;
+	}
+
+	memcpy(*dest, src, len);
+	return 0;
 }
 
 char *dstrprintf(const char *format, ...) {
@@ -184,7 +216,7 @@ int dstrvcatf(char **str, const char *format, va_list args) {
 		tail = *str + len;
 		ret = vsnprintf(tail, tail_len + 1, format, copy);
 		if (ret < 0 || (size_t)ret != tail_len) {
-			assert(!"Length of formatted string changed");
+			bfs_bug("Length of formatted string changed");
 			goto fail;
 		}
 	}
@@ -200,6 +232,28 @@ fail:
 	return -1;
 }
 
+int dstrescat(char **dest, const char *str, enum wesc_flags flags) {
+	return dstrnescat(dest, str, SIZE_MAX, flags);
+}
+
+int dstrnescat(char **dest, const char *str, size_t n, enum wesc_flags flags) {
+	size_t len = *dest ? dstrlen(*dest) : 0;
+
+	// Worst case growth is `ccc...` => $'\xCC\xCC\xCC...'
+	n = strnlen(str, n);
+	size_t cap = len + 4 * n + 3;
+	if (dstreserve(dest, cap) != 0) {
+		return -1;
+	}
+
+	char *cur = *dest + len;
+	char *end = *dest + cap + 1;
+	cur = wordnesc(cur, end, str, n, flags);
+	bfs_assert(cur != end, "wordesc() result truncated");
+
+	return dstresize(dest, cur - *dest);
+}
+
 void dstrfree(char *dstr) {
 	if (dstr) {
 		free(dstrheader(dstr));
