@@ -2,26 +2,29 @@
 // SPDX-License-Identifier: 0BSD
 
 #include "stat.h"
+#include "atomic.h"
 #include "bfstd.h"
 #include "config.h"
-#include <assert.h>
+#include "diag.h"
+#include "sanity.h"
 #include <errno.h>
 #include <fcntl.h>
-#include <stdbool.h>
 #include <string.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 
 #if defined(STATX_BASIC_STATS) && (!__ANDROID__ || __ANDROID_API__ >= 30)
-#	define BFS_LIBC_STATX true
+#  define BFS_HAS_LIBC_STATX true
 #elif __linux__
-#	include <linux/stat.h>
-#	include <sys/syscall.h>
-#	include <unistd.h>
+#  include <linux/stat.h>
+#  include <sys/syscall.h>
+#  include <unistd.h>
 #endif
 
-#if BFS_LIBC_STATX || defined(__NR_statx)
-#	define BFS_STATX true
+#ifndef BFS_USE_STATX
+#  if BFS_HAS_LIBC_STATX || defined(SYS_statx)
+#    define BFS_USE_STATX true
+#  endif
 #endif
 
 const char *bfs_stat_field_name(enum bfs_stat_field field) {
@@ -58,7 +61,7 @@ const char *bfs_stat_field_name(enum bfs_stat_field field) {
 		return "modification time";
 	}
 
-	assert(!"Unrecognized stat field");
+	bfs_bug("Unrecognized stat field");
 	return "???";
 }
 
@@ -127,23 +130,24 @@ static int bfs_stat_impl(int at_fd, const char *at_path, int at_flags, struct bf
 	return ret;
 }
 
-#if BFS_STATX
+#if BFS_USE_STATX
 
 /**
  * Wrapper for the statx() system call, which had no glibc wrapper prior to 2.28.
  */
 static int bfs_statx(int at_fd, const char *at_path, int at_flags, unsigned int mask, struct statx *buf) {
-#if __has_feature(memory_sanitizer)
-	// -fsanitize=memory doesn't know about statx(), so tell it the memory
-	// got initialized
-	memset(buf, 0, sizeof(*buf));
+#if BFS_HAS_LIBC_STATX
+	int ret = statx(at_fd, at_path, at_flags, mask, buf);
+#else
+	int ret = syscall(SYS_statx, at_fd, at_path, at_flags, mask, buf);
 #endif
 
-#if BFS_LIBC_STATX
-	return statx(at_fd, at_path, at_flags, mask, buf);
-#else
-	return syscall(__NR_statx, at_fd, at_path, at_flags, mask, buf);
-#endif
+	if (ret == 0) {
+		// -fsanitize=memory doesn't know about statx()
+		sanitize_init(buf);
+	}
+
+	return ret;
 }
 
 /**
@@ -240,21 +244,21 @@ static int bfs_statx_impl(int at_fd, const char *at_path, int at_flags, struct b
 	return ret;
 }
 
-#endif // BFS_STATX
+#endif // BFS_USE_STATX
 
 /**
  * Calls the stat() implementation with explicit flags.
  */
 static int bfs_stat_explicit(int at_fd, const char *at_path, int at_flags, int x_flags, struct bfs_stat *buf) {
-#if BFS_STATX
-	static bool has_statx = true;
+#if BFS_USE_STATX
+	static atomic bool has_statx = true;
 
-	if (has_statx) {
+	if (load(&has_statx, relaxed)) {
 		int ret = bfs_statx_impl(at_fd, at_path, at_flags | x_flags, buf);
 		// EPERM is commonly returned in a seccomp() sandbox that does
 		// not allow statx()
 		if (ret != 0 && (errno == ENOSYS || errno == EPERM)) {
-			has_statx = false;
+			store(&has_statx, false, relaxed);
 		} else {
 			return ret;
 		}
@@ -304,12 +308,12 @@ int bfs_stat(int at_fd, const char *at_path, enum bfs_stat_flags flags, struct b
 
 	// Check __GNU__ to work around https://lists.gnu.org/archive/html/bug-hurd/2021-12/msg00001.html
 #if defined(AT_EMPTY_PATH) && !__GNU__
-	static bool has_at_ep = true;
-	if (has_at_ep) {
+	static atomic bool has_at_ep = true;
+	if (load(&has_at_ep, relaxed)) {
 		at_flags |= AT_EMPTY_PATH;
 		int ret = bfs_stat_explicit(at_fd, "", at_flags, x_flags, buf);
 		if (ret != 0 && errno == EINVAL) {
-			has_at_ep = false;
+			store(&has_at_ep, false, relaxed);
 		} else {
 			return ret;
 		}
@@ -341,7 +345,7 @@ const struct timespec *bfs_stat_time(const struct bfs_stat *buf, enum bfs_stat_f
 	case BFS_STAT_MTIME:
 		return &buf->mtime;
 	default:
-		assert(!"Invalid stat field for time");
+		bfs_bug("Invalid stat field for time");
 		errno = EINVAL;
 		return NULL;
 	}
