@@ -24,7 +24,7 @@ enum bfs_spawn_flags {
  */
 struct bfs_spawn {
 	enum bfs_spawn_flags flags;
-	struct bfs_spawn_action *actions;
+	struct bfs_spawn_action *head;
 	struct bfs_spawn_action **tail;
 };
 
