@@ -9,11 +9,12 @@ umask 022
 export LC_ALL=C
 export TZ=UTC0
 
-export ASAN_OPTIONS="abort_on_error=1:log_to_syslog=0"
-export LSAN_OPTIONS="abort_on_error=1:log_to_syslog=0"
-export MSAN_OPTIONS="abort_on_error=1:log_to_syslog=0"
-export TSAN_OPTIONS="abort_on_error=1:log_to_syslog=0"
-export UBSAN_OPTIONS="abort_on_error=1:log_to_syslog=0"
+SAN_OPTIONS="halt_on_error=1:log_to_syslog=0"
+export ASAN_OPTIONS="$SAN_OPTIONS"
+export LSAN_OPTIONS="$SAN_OPTIONS"
+export MSAN_OPTIONS="$SAN_OPTIONS"
+export TSAN_OPTIONS="$SAN_OPTIONS"
+export UBSAN_OPTIONS="$SAN_OPTIONS"
 
 export LS_COLORS=""
 unset BFS_COLORS
@@ -412,7 +413,9 @@ make_deep "$TMP/deep"
 function make_rainbow() {
     "$XTOUCH" -p "$1/file.txt"
     "$XTOUCH" -p "$1/file.dat"
-    "$XTOUCH" -p "$1/star".{gz,tar,tar.gz}
+    "$XTOUCH" -p "$1/lower".{gz,tar,tar.gz}
+    "$XTOUCH" -p "$1/upper".{GZ,TAR,TAR.GZ}
+    "$XTOUCH" -p "$1/lu.tar.GZ" "$1/ul.TAR.gz"
     ln -s file.txt "$1/link.txt"
     "$XTOUCH" -p "$1/mh1"
     ln "$1/mh1" "$1/mh2"
@@ -429,6 +432,7 @@ function make_rainbow() {
     chmod +t "$1"/sticky*
     "$XTOUCH" -p "$1"/exec.sh
     chmod +x "$1"/exec.sh
+    "$XTOUCH" -p "$1/"$'\e[1m/\e[0m'
 }
 make_rainbow "$TMP/rainbow"
 
